/* $Id$ */
#include <lcms.h>
#include "common.h"
#include "cmsprofile.h"
#include "cmsimage.h"
#include "module.h"

struct cms_profile {
  cmsHPROFILE profile;
};

static cmsHPROFILE get_profile_CIExyY(LPcmsCIExyY white, LPcmsCIExyYTRIPLE primaries, double gamma);
static CmsProfile *cmsProfile_alloc(void);
static void cmsProfile_free(CmsProfile *profile);

static void copy_alpha_channel(unsigned char *pdst, const unsigned char *psrc, ImageType type, int len);
static cmsHTRANSFORM create_transform(CmsProfile *profile_in, ImageType type_in, CmsProfile *profile_out, ImageType type_out);
static DWORD image_dw_type(Image *image);
static DWORD imageType_dw_type(ImageType type);
static void transform_data(cmsHTRANSFORM transform, Image *image_in, Image *image_out);
static void transform_indexed_colors(cmsHTRANSFORM transform, Image *image_in, Image *image_out);
#if defined DEBUG
static void info_print(cmsHPROFILE profile);
#endif /* DEBUG */

CmsProfile *cmsProfile_open_mem(void *ptr, size_t len)
{
  CmsProfile *profile;
  DEBUG_PRINT("open_mem()\n");
  profile = cmsProfile_alloc();
  if (ptr != NULL) {
    profile->profile = cmsOpenProfileFromMem(ptr, len);
#if defined DEBUG
    info_print(profile->profile);
#endif /* DEBUG */
  }
  return profile;
}

CmsProfile *cmsProfile_open_file(const char *filename)
{
  CmsProfile *profile;
  DEBUG_PRINT("open_file() filename = '%s'\n", filename);
  profile = cmsProfile_alloc();
  if (filename != NULL) {
    profile->profile = cmsOpenProfileFromFile(filename, "r");
#if defined DEBUG
    info_print(profile->profile);
#endif /* DEBUG */
  }
  return profile;
}

CmsProfile *cmsProfile_open_sRGB(void)
{
  CmsProfile *profile;
  DEBUG_PRINT("open_sRGB()\n");
  profile = cmsProfile_alloc();
  profile->profile = cmsCreate_sRGBProfile();
#if defined DEBUG
  info_print(profile->profile);
#endif /* DEBUG */
  return profile;
}

CmsProfile *cmsProfile_open_chromaticities(double white_x, double white_y,
                                           double red_x, double red_y,
                                           double green_x, double green_y,
                                           double blue_x, double blue_y,
                                           double gamma_of_file)
{
  CmsProfile *profile;
  cmsCIExyY white;
  cmsCIExyYTRIPLE primaries;
  DEBUG_PRINT("open_chromaticities()\n");
  profile = cmsProfile_alloc();
  white.x = white_x;
  white.y = white_y;
  primaries.Red.x = red_x;
  primaries.Red.y = red_y;
  primaries.Green.x = green_x;
  primaries.Green.y = green_y;
  primaries.Blue.x = blue_x;
  primaries.Blue.y = blue_y;
  white.Y = primaries.Red.Y = primaries.Green.Y = primaries.Blue.Y = 1.0;
  profile->profile = get_profile_CIExyY(&white, &primaries, gamma_of_file);
#if defined DEBUG
  info_print(profile->profile);
#endif /* DEBUG */
  return profile;
}

#if 0
CmsProfile *cmsProfile_open_chromaticities(double white_x, double white_y,
                                           double red_x, double red_y,
                                           double green_x, double green_y,
                                           double blue_x, double blue_y,
                                           unsigned short *gamma_red,
                                           unsigned short *gamma_green,
                                           unsigned short *gamma_blue)
{
  cmsHPROFILE profile;
  LPGAMMATABLE tables[3];

  for (int i = 0; i < 3; i++) {
    tables[i] = cmsAllocGamma(256);
  }
  CopyMemory(tables[0]->GammaTable, gamma_red, 256*sizeof(WORD));
  CopyMemory(tables[1]->GammaTable, gamma_green, 256*sizeof(WORD));
  CopyMemory(tables[2]->GammaTable, gamma_blue, 256*sizeof(WORD));
  profile = cmsCreateRGBProfile(white, primaries, tables);
  for (int i = 0; i < 3; i++) {
    cmsFreeGamma(tables[i]);
  }
  return profile;
}
#endif /* 0 */

CmsProfile *cmsProfile_open_temp(int tempK, double gamma_of_file, int rgb_profile)
{
  CmsProfile *profile;
  cmsCIExyY white;
  cmsCIExyYTRIPLE primaries = {
    { 0.6400, 0.3300, 1.0 },
    { 0.3000, 0.6000, 1.0 },
    { 0.1500, 0.0600, 1.0 }
  };
  LPcmsCIExyYTRIPLE pp;

  cmsWhitePointFromTemp(tempK, &white);
  pp = (rgb_profile) ? &primaries : NULL;
  profile = cmsProfile_alloc();
  profile->profile = get_profile_CIExyY(&white, pp, gamma_of_file);
#if defined DEBUG
  info_print(profile->profile);
#endif /* DEBUG */
  return profile;
}

int cmsProfile_close(CmsProfile *profile)
{
  int rc = FALSE;
  if (profile == NULL) {
    return rc;
  }
  if (cmsCloseProfile(profile->profile)) {
    profile->profile = NULL;
    rc = TRUE;
  }
  cmsProfile_free(profile);
  return rc;
}

static cmsHPROFILE get_profile_CIExyY(LPcmsCIExyY white, LPcmsCIExyYTRIPLE primaries, double gamma_of_file)
{
  cmsHPROFILE profile;
  LPGAMMATABLE trans;

  trans = cmsBuildGamma(256, 1.0/gamma_of_file);
  if (trans == NULL) {
    return NULL;
  }
  if (primaries != NULL) {
    LPGAMMATABLE tables[3];
    tables[0] = tables[1] = tables[2] = trans;
    profile = cmsCreateRGBProfile(white, primaries, tables);
  } else {
    profile = cmsCreateGrayProfile(white, trans);
  }
  cmsFreeGamma(trans);
  return profile;
}

static CmsProfile *cmsProfile_alloc(void)
{
  CmsProfile *p;
  p = XMALLOC(CmsProfile, 1);
  p->profile = NULL;
  return p;
}

static void cmsProfile_free(CmsProfile *profile)
{
  XFREE(profile);
}


Image *image_transform_new(Image *image_in, CmsProfile *profile_in, CmsProfile *profile_out)
{
  Image *image_out;
  unsigned char *data;
  ImageType type;
  int bpp; /* bytes per pixel */
  int w, h, xoff, yoff;

  image_get_size(image_in, &w, &h);
  image_get_offset(image_in, &xoff, &yoff);
  type = image_get_type(image_in);
  image_out = image_new();
  switch (type) {
  case IMAGE_TYPE_GRAYA_8:
    image_set_type(image_out, IMAGE_TYPE_RGBA_8);
    bpp = imageType_bytes_per_pixel(IMAGE_TYPE_RGBA_8);
    break;
  case IMAGE_TYPE_GRAYA_16:
    image_set_type(image_out, IMAGE_TYPE_RGBA_16);
    bpp = imageType_bytes_per_pixel(IMAGE_TYPE_RGBA_16);
    break;
  default:
    image_set_type(image_out, type);
    bpp = imageType_bytes_per_pixel(type);
    break;
  }
  data = XMALLOC(unsigned char, w*h*bpp);
  copy_alpha_channel(data, image_get_data(image_in), type, w*h);
  image_set_data(image_out, data);
  image_set_size(image_out, w, h);
  image_set_offset(image_out, xoff, yoff);
  if (type == IMAGE_TYPE_INDEXED_RGBA_8) {
    image_transform_palette(image_in, profile_in, image_out, profile_out);
  } else {
    image_transform_data(image_in, profile_in, image_out, profile_out);
  }
  return image_out;
}

static void copy_alpha_channel(unsigned char *pdst, const unsigned char *psrc, ImageType type, int len)
{
  switch (type) {
  case IMAGE_TYPE_GRAYA_8: /* to RGBA_8 */
    for (int i = 0; i < len; i++) {
      pdst+=3;
      psrc++;
      *(pdst++) = *(psrc++);
    }
    break;
  case IMAGE_TYPE_GRAYA_16: /* to RGBA_16 */
    for (int i = 0; i < len; i++) {
      pdst+=6;
      psrc+=2;
      *(pdst++) = *(psrc++);
      *(pdst++) = *(psrc++);
    }
    break;
  case IMAGE_TYPE_RGBA_8:
    for (int i = 0; i < len; i++) {
      pdst+=3;
      psrc+=3;
      *(pdst++) = *(psrc++);
    }
    break;
  case IMAGE_TYPE_RGBA_16:
    for (int i = 0; i < len; i++) {
      pdst+=6;
      psrc+=6;
      *(pdst++) = *(psrc++);
      *(pdst++) = *(psrc++);
    }
    break;
  default:
    break;
  }
}

void image_transform_data(Image *image_in, CmsProfile *profile_in, Image *image_out, CmsProfile *profile_out)
{
  cmsHTRANSFORM transform;
  ImageType type_in, type_out;

  type_in = image_get_type(image_in);
  type_out = image_get_type(image_out);
  transform = create_transform(profile_in, type_in, profile_out, type_out);
  if (transform == NULL) {
    fprintf(stderr, "Cannot transform data.\n");
    return;
  }
  transform_data(transform, image_in, image_out);
  cmsDeleteTransform(transform);
}

void image_transform_palette(Image *image_in, CmsProfile *profile_in, Image *image_out, CmsProfile *profile_out)
{
  cmsHTRANSFORM transform;
  ImageType type_in, type_out;

  type_in = image_get_type(image_in);
  type_out = image_get_type(image_out);
  transform = create_transform(profile_in, type_in, profile_out, type_out);
  if (transform == NULL) {
    fprintf(stderr, "Cannot transform palette.\n");
    return;
  }
  transform_indexed_colors(transform, image_in, image_out);
  cmsDeleteTransform(transform);
}

void cms_data_transform(unsigned char *data_in, ImageType type_in, CmsProfile *profile_in, unsigned char *data_out, ImageType type_out, CmsProfile *profile_out, unsigned int len)
{
  cmsHTRANSFORM transform;

  transform = create_transform(profile_in, type_in, profile_out, type_out);
  if (transform == NULL) {
    fprintf(stderr, "Cannot transform data.\n");
    return;
  }
  cmsDoTransform(transform, data_in, data_out, len);
  cmsDeleteTransform(transform);
}

static cmsHTRANSFORM create_transform(CmsProfile *profile_in, ImageType type_in, CmsProfile *profile_out, ImageType type_out)
{
  cmsHPROFILE pin, pout;
  cmsHTRANSFORM transform;
  int intent = INTENT_PERCEPTUAL;
  DWORD dw_flags = 0;
  DWORD dw_type_in, dw_type_out;

  pin = profile_in->profile;
  pout = profile_out->profile;
  dw_type_in = imageType_dw_type(type_in);
  dw_type_out = imageType_dw_type(type_out);
  transform = cmsCreateTransform(pin, dw_type_in, pout, dw_type_out,
                                 intent, dw_flags);
  return transform;
}

static DWORD image_dw_type(Image *image)
{
  ImageType type;

  type = image_get_type(image);
  return imageType_dw_type(type);
}

static DWORD imageType_dw_type(ImageType type)
{
  DWORD dw_type;

  switch (type) {
  case IMAGE_TYPE_INDEXED_GRAY_8:
    dw_type = TYPE_GRAY_8;
    DEBUG_PRINT("dw_type=%s\n", "GRAY_8(INDEX)");
    break;
  case IMAGE_TYPE_INDEXED_GRAYA_8:
    dw_type = TYPE_GRAYA_8;
    DEBUG_PRINT("dw_type=%s\n", "GRAYA_8(INDEX)");
    break;
  case IMAGE_TYPE_INDEXED_RGB_8:
    dw_type = TYPE_RGB_8;
    DEBUG_PRINT("dw_type=%s\n", "RGB_8(INDEX)");
    break;
  case IMAGE_TYPE_INDEXED_RGBA_8:
    dw_type = TYPE_RGBA_8;
    DEBUG_PRINT("dw_type=%s\n", "RGBA_8(INDEX)");
    break;
  case IMAGE_TYPE_GRAY_8:
    dw_type = TYPE_GRAY_8;
    DEBUG_PRINT("dw_type=%s\n", "GRAY_8");
    break;
  case IMAGE_TYPE_GRAY_16:
    dw_type = TYPE_GRAY_16_SE;
    DEBUG_PRINT("dw_type=%s\n", "GRAY_16_SE");
    break;
  case IMAGE_TYPE_GRAYA_8:
    dw_type = TYPE_GRAYA_8;
    DEBUG_PRINT("dw_type=%s\n", "GRAYA_8");
    break;
  case IMAGE_TYPE_GRAYA_16:
    dw_type = TYPE_GRAYA_16_SE;
    DEBUG_PRINT("dw_type=%s\n", "GRAYA_16_SE");
    break;
  case IMAGE_TYPE_RGB_8:
    dw_type = TYPE_RGB_8;
    DEBUG_PRINT("dw_type=%s\n", "RGB_8");
    break;
  case IMAGE_TYPE_RGBA_8:
    dw_type = TYPE_RGBA_8;
    DEBUG_PRINT("dw_type=%s\n", "RGBA_8");
    break;
  case IMAGE_TYPE_RGB_16:
    dw_type = TYPE_RGB_16_SE;
    DEBUG_PRINT("dw_type=%s\n", "RGB_16_SE");
    break;
  case IMAGE_TYPE_RGBA_16:
    dw_type = TYPE_RGBA_16_SE;
    DEBUG_PRINT("dw_type=%s\n", "RGBA_16_SE");
    break;
  case IMAGE_TYPE_CMYK_8:
    dw_type = TYPE_CMYK_8;
    DEBUG_PRINT("dw_type=%s\n", "CMYK_8");
    break;
  default:
    DEBUG_PRINT("dw_type=%s\n", "unknown");
    break;
  }
  return dw_type;
}

static void transform_data(cmsHTRANSFORM transform, Image *image_in, Image *image_out)
{
  unsigned char *data_in, *data_out;
  unsigned char *p_in, *p_out;
  int bpp_in, bpp_out;
  int w_in, h_in, w_out, h_out;
  int xoff_in, yoff_in, xoff_out, yoff_out;

  DEBUG_PRINT("transform_data() --- start.\n");
  image_get_size(image_in, &w_in, &h_in);
  image_get_offset(image_in, &xoff_in, &yoff_in);
  bpp_in = imageType_bytes_per_pixel(image_get_type(image_in));
  if ((bpp_in <= 0) || (w_in <= 0) || (h_in <= 0)) {
    DEBUG_PRINT("bpp_in = %d, w_in = %d, h_in = %d\n", bpp_in, w_in, h_in);
    return;
  }
  DEBUG_PRINT("transform_data(): in check OK.\n");
  image_get_size(image_out, &w_out, &h_out);
  image_get_offset(image_out, &xoff_out, &yoff_out);
  bpp_out = imageType_bytes_per_pixel(image_get_type(image_out));
  if ((bpp_out <= 0) || (w_out <= 0) || (h_out <= 0)) {
    DEBUG_PRINT("bpp_out = %d, w_out = %d, h_out = %d\n", bpp_out, w_out, h_out);
    return;
  }
  DEBUG_PRINT("transform_data(): out check OK.\n");
  if ((w_in > w_out) || (h_in > h_out)) {
    DEBUG_PRINT("w_in = %d, w_out = %d, h_in = %d, h_out = %d\n", w_in, w_out, h_in, h_out);
    return;
  }
  DEBUG_PRINT("transform_data(): check OK.\n");
  data_in = image_get_data(image_in);
  data_out = image_get_data(image_out);
  p_in = data_in;
  p_out = data_out+(w_out*yoff_in+xoff_in)*bpp_out;
  if ((xoff_in == 0) && (xoff_out == 0) && (w_in == w_out)) {
    cmsDoTransform(transform, p_in, p_out, w_in*h_in);
  } else {
    for (int j = 0; j < h_in; j++) {
      cmsDoTransform(transform, p_in, p_out, w_in);
      p_in += w_in*bpp_in;
      p_out += w_out*bpp_out;
    }
  }
  DEBUG_PRINT("transform_data() --- end.\n");
}

static void transform_indexed_colors(cmsHTRANSFORM transform, Image *image_in, Image *image_out)
{
  ImageType type;
  int num;

  type = image_get_type(image_in);
  num = image_get_num_indexes(image_in);
  image_set_num_indexes(image_out, num);
  for (int i = 0; i < num; i++) {
    unsigned char din[4];
    unsigned char dout[4];

    image_get_index(image_in, i, din);
    switch (type) {
    case IMAGE_TYPE_INDEXED_GRAYA_8:
      copy_alpha_channel(dout, din, IMAGE_TYPE_GRAYA_8, 1);
      break;
    case IMAGE_TYPE_INDEXED_RGBA_8:
      copy_alpha_channel(dout, din, IMAGE_TYPE_RGBA_8, 1);
      break;
    default:
      dout[3] = 0xff;
      break;
    }
    cmsDoTransform(transform, din, dout, 1);
    image_set_index(image_out, i, dout);
  }
}

#if defined DEBUG
static void info_print(cmsHPROFILE profile)
{
  FILE *fp;

  fp = stderr;
  fprintf(fp, "product name: %s\n", cmsTakeProductName(profile));
  fprintf(fp, "product desc: %s\n", cmsTakeProductDesc(profile));
  fprintf(fp, "manufacturer: %s\n", cmsTakeManufacturer(profile));
  fprintf(fp, "model: %s\n", cmsTakeModel(profile));
  fprintf(fp, "copyright: %s\n", cmsTakeCopyright(profile));
  fprintf(fp, "ICC version: %x\n", cmsGetProfileICCversion(profile));
  fprintf(fp, "ICC header flags: %x\n", cmsTakeHeaderFlags(profile));
  fprintf(fp, "ICC header attributes: %x\n", cmsTakeHeaderAttributes(profile));
  fprintf(fp, "intent field of the header: %x\n", cmsTakeRenderingIntent(profile));
  fprintf(fp, "number of tags: %d\n", cmsGetTagCount(profile));
  fprintf(fp, "\n");
}
#endif /* DEBUG */
