/* $Id$ */
#include "common.h"
#include "module.h"
#include "image_utils.h"
#include "cmsprofile.h"
#include "cmsimage.h"

#define NUM_FRAMES 3
#define BGDATA_LEN 8 /* RGBA_16 */
#define BGUSER_LEN 6 /* RGB_16 */

#define IMAGE_TYPE_IS_INDEXED(type)          \
  (((type) == IMAGE_TYPE_INDEXED_GRAY_8)  || \
   ((type) == IMAGE_TYPE_INDEXED_GRAYA_8) || \
   ((type) == IMAGE_TYPE_INDEXED_RGB_8)   || \
   ((type) == IMAGE_TYPE_INDEXED_RGBA_8))

struct imageWork {
  ImageType type;
  Image *current;
  Image *previous;
  Image *frame[NUM_FRAMES];
  int num_frames;
  int last_index;
  BgPrefer bg_prefer;
  ImageType bgtype;
  unsigned char *bgdata;
  unsigned char *bguser;
#if 0
  unsigned char *bg_previous;
#endif /* 0 */
  int dispose_op;
  int updated_xoff;
  int updated_yoff;
  int updated_width;
  int updated_height;
  CmsProfile *profile_embedded;
  CmsProfile *profile_in;
  CmsProfile *profile_out;
  int use_infile;
};

static ImageType get_imageWork_type(ImageSet *images);
static int get_image_background(ImageWork *work, Image *image, unsigned char *pbgdata);
static CmsProfile *imageWork_get_profile_in(ImageWork *work);
static void image_copy_data(Image *dst, Image *src);
static void image_copy_palette(Image *dst, Image *src);
static void image_fill_pixel(Image *image, unsigned char *pixel);
static void image_blend(Image *image, Image *newframe,
                        CmsProfile *profile_in, CmsProfile *profile_out,
                        unsigned char *bgdata);
static void image_blend_index_to_rgb8(Image *image, Image *newframe, const unsigned char *bgdata);
static void image_blend_rgb(Image *image, Image *newframe, const unsigned char *bgdata);
static void copy_gray8_to_rgb8(unsigned char *pdst, const unsigned char *psrc, int len);
static void copy_gray8_to_rgba8(unsigned char *pdst, const unsigned char *psrc, int len);
static void copy_gray16_to_rgb16(unsigned char *pdst, const unsigned char *psrc, int len);
static void copy_graya8_to_rgba8(unsigned char *pdst, const unsigned char *psrc, int len);
static void copy_graya16_to_rgba16(unsigned char *pdst, const unsigned char *psrc, int len);
static void copy_rgb8_to_rgba8(unsigned char *pdst, const unsigned char *psrc, int len);
static void copy_cmyk8_to_rgb8(unsigned char *pdst, const unsigned char *psrc, int len);
static void blend_c8(unsigned char *p1, const unsigned char *p2, const unsigned char *bg, int num);
static void blend_c16(unsigned char *p1, const unsigned char *p2, const unsigned char *bg, int num);
static void get_rgb8(unsigned char *rgb, unsigned short *r, unsigned short *g, unsigned short *b);
static void get_rgb16(unsigned char *rgb, unsigned short *r, unsigned short *g, unsigned short *b);
static void set_rgb16(unsigned char *rgb, unsigned short r, unsigned short g, unsigned short b);
static unsigned short get_c16(unsigned char *c16);
static void set_c16(unsigned char *c16, unsigned short c);
static unsigned short get_c8(unsigned char *c8);


void imageWork_init(ImageWork *work, ImageSet *images)
{
  ImageType type;
  size_t size;
  int width, height;

  type = get_imageWork_type(images);
  imageSet_get_size(images, &width, &height);
  size = imageType_bytes_per_pixel(type)*width*height;
  for (int i = 0; i < work->num_frames; i++) {
    Image *image;
    size_t size0;
    int w, h;
    image = work->frame[i];
    image_get_size(image, &w, &h);
    size0 = imageType_bytes_per_pixel(work->type)*w*h;
    if (size != size0) {
      unsigned char *data;
      data = image_get_data(image);
      if (data != NULL) {
        XFREE(data);
      }
      data = XMALLOC(unsigned char, size);
      image_set_data(image, data);
    }
    image_set_type(image, type);
    image_set_size(image, width, height);
    image_set_dispose_op(image, DISPOSE_OP_BACKGROUND);
  }
  work->type = type;
  work->current = work->frame[0];
  work->previous = NULL;
  work->last_index = 2;

  imageSet_get_background(images, work->bgdata, &(work->bgtype));
  work->profile_embedded = imageSet_get_profile(images);
#if defined DEBUG
  DEBUG_PRINT(" bg_prefer = %d\n", work->bg_prefer);
  DEBUG_PRINT(" bgtype = %d\n", work->bgtype);
  for (int i = 0; i < 8; i++) {
    DEBUG_PRINT(" bgdata[%d] = %x\n", i, work->bgdata[i]);
  }
  for (int i = 0; i < 6; i++) {
    DEBUG_PRINT(" bguser[%d] = %x\n", i, work->bguser[i]);
  }
#endif /* DEBUG */
}

void imageWork_get_background(ImageWork *work, unsigned short *r, unsigned short *g, unsigned short *b)
{
  BgPrefer bg_prefer;
  Image *image;
  unsigned char bgdata[BGDATA_LEN];

  DEBUG_PRINT(" bg_prefer=%d bgtype=%d\n", work->bg_prefer, work->bgtype);
  bg_prefer = work->bg_prefer;
  if (bg_prefer == BG_USER) {
    get_rgb16(work->bguser, r, g, b);
    return;
  }
  /* otherwise, bg_prefer = BG_IMAGE */
  image = work->current;
  if (get_image_background(work, image, bgdata)) {
    unsigned char bgtmp[4];
    switch (image_get_type(image)) {
    case IMAGE_TYPE_INDEXED_RGB_8:
    case IMAGE_TYPE_INDEXED_RGBA_8:
      image_get_index(image, bgdata[0], bgtmp);
      get_rgb8(bgtmp, r, g, b);
      break;
    case IMAGE_TYPE_RGB_8:
      get_rgb8(bgdata, r, g, b);
      break;
    case IMAGE_TYPE_RGB_16:
      get_rgb16(bgdata, r, g, b);
      break;
    }
  }
}

void imageWork_set_background(ImageWork *work, unsigned short r, unsigned short g, unsigned short b, BgPrefer prefer)
{
  work->bg_prefer = prefer;
  set_rgb16(work->bguser, r, g, b);
}

static ImageType get_imageWork_type(ImageSet *images)
{
  ImageType type, worktype;
  Image *image;
  int num;

  image = imageSet_get_first(images);
  if (image == NULL) {
    return IMAGE_TYPE_UNKNOWN;
  }
  type = image_get_type(image);
  switch (type) {
  case IMAGE_TYPE_INDEXED_GRAY_8:
  case IMAGE_TYPE_INDEXED_GRAYA_8:
  case IMAGE_TYPE_INDEXED_RGB_8:
  case IMAGE_TYPE_INDEXED_RGBA_8:
    worktype = IMAGE_TYPE_INDEXED_RGB_8;
    num = imageSet_get_num_frames(images);
    for (int j = 0; j < num; j++) {
      int n = image_get_num_indexes(image);
      for (int i = 0; i < n; i++) {
        unsigned char data[4];
        unsigned char alpha = 0xff;
        type = image_get_type(image);
        image_get_index(image, i, data);
        if (type == IMAGE_TYPE_INDEXED_GRAYA_8) {
          worktype = IMAGE_TYPE_INDEXED_RGBA_8;
          alpha = data[1];
        } else if (type == IMAGE_TYPE_INDEXED_RGBA_8) {
          worktype = IMAGE_TYPE_INDEXED_RGBA_8;
          alpha = data[3];
        }
        if ((alpha > 0x0) && (alpha < 0xff)) {
          worktype = IMAGE_TYPE_RGB_8;
          DEBUG_PRINT("imageWork: type = %d\n", worktype);
          return worktype;
        }
      }
      image = image_next(image);
    }
    break;
  case IMAGE_TYPE_GRAY_8:
  case IMAGE_TYPE_GRAYA_8:
  case IMAGE_TYPE_RGB_8:
  case IMAGE_TYPE_RGBA_8:
  case IMAGE_TYPE_CMYK_8:
    worktype = IMAGE_TYPE_RGB_8;
    break;
  case IMAGE_TYPE_GRAY_16:
  case IMAGE_TYPE_GRAYA_16:
  case IMAGE_TYPE_RGB_16:
  case IMAGE_TYPE_RGBA_16:
    worktype = IMAGE_TYPE_RGB_16;
    break;
  default:
    break;
  }
  DEBUG_PRINT("imageWork: type = %d\n", worktype);
  return worktype;
}

ImageWork *imageWork_alloc(void)
{
  ImageWork *work;

  work = XMALLOC(ImageWork, 1);
  work->type = IMAGE_TYPE_UNKNOWN;
  work->num_frames = NUM_FRAMES;
  for (int i = 0; i < work->num_frames; i++) {
    Image *image;
    image = image_new();
    image_set_dispose_op(image, DISPOSE_OP_BACKGROUND);
    work->frame[i] = image;
  }
  work->current = work->frame[0];
  work->previous = NULL;
  work->last_index = 2;
  work->bg_prefer = BG_IMAGE;
  work->bgtype = IMAGE_TYPE_UNKNOWN;
  work->bgdata = XMALLOC(unsigned char, BGDATA_LEN);
  work->bguser = XMALLOC(unsigned char, BGUSER_LEN);
  memset(work->bgdata, 0x7f, BGDATA_LEN);
  set_rgb16(work->bguser, 0x7f7f, 0x7f7f, 0x7f7f);
  work->updated_xoff = 0;
  work->updated_yoff = 0;
  work->updated_width = 0;
  work->updated_height = 0;
  work->profile_embedded = NULL;
  work->profile_in = NULL;
  work->profile_out = NULL;
  work->use_infile = FALSE;
  return work;
}

void imageWork_free(ImageWork *work)
{
  for (int i = 0; i < work->num_frames; i++) {
    image_free(work->frame[i]);
  }
  XFREE(work);
}

void imageWork_update(ImageWork *work, Image *newframe)
{
  Image *image;
  Image *last_frame;
  ImageType newtype;
  int dispose_op0;
  int dispose_op;
  int blend_op;
  unsigned char bgdata[BGDATA_LEN];
  ImageType bgtype;
  int w, h, xoff, yoff;
  int full_image = FALSE;
  CmsProfile *profile_in;
  CmsProfile *profile_out;

  newtype = image_get_type(newframe);
  DEBUG_PRINT(" image type: work = %d / new = %d\n", work->type, newtype);
  last_frame = NULL;
  image = work->current;
  image_set_type(image, work->type);
  dispose_op = image_get_dispose_op(newframe);
  if (dispose_op == DISPOSE_OP_PREVIOUS) {
    work->last_index = (work->last_index == 2) ? 1 : 2;
    last_frame = work->frame[work->last_index];
    image_copy_palette(last_frame, image);
    image_copy_data(last_frame, image);
  }
  /* dispose operation */
  dispose_op0 = image_get_dispose_op(image);
  DEBUG_PRINT(" dispose_op type: %d\n", dispose_op0);
  switch (dispose_op0) {
  case DISPOSE_OP_NONE:
    break;
  case DISPOSE_OP_BACKGROUND:
    get_image_background(work, image, bgdata);
    image_fill_pixel(image, bgdata);
    full_image = TRUE;
    break;
  case DISPOSE_OP_PREVIOUS:
    image_copy_palette(image, work->previous);
    image_copy_data(image, work->previous);
    full_image = TRUE;
    break;
  default:
    DEBUG_PRINT(" unknown dipose_op type %d. why?\n", dispose_op0);
    break;
  }
  if (full_image) {
    image_get_size(image, &w, &h);
    xoff = 0;
    yoff = 0;
  } else {
    image_get_size(newframe, &w, &h);
    image_get_offset(newframe, &xoff, &yoff);
  }
  work->updated_xoff = xoff;
  work->updated_yoff = yoff;
  work->updated_width = w;
  work->updated_height = h;
  if (work->type == IMAGE_TYPE_INDEXED_RGBA_8) {
    work->dispose_op = dispose_op0;
  } else {
    work->dispose_op = DISPOSE_OP_NONE;
  }
  image_set_dispose_op(image, dispose_op);
  profile_in = imageWork_get_profile_in(work);
  profile_out = work->profile_out;
  if ((profile_in != NULL) && (profile_out != NULL)) {
    image_transform_palette(newframe, profile_in, image, profile_out);
  } else {
    image_copy_palette(image, newframe);
  }
  work->previous = last_frame;
  /* blend operation */
  blend_op = image_get_blend_op(newframe);
  DEBUG_PRINT(" blend_op type: %d\n", blend_op);
  switch (blend_op) {
  case BLEND_OP_NONE:
    get_image_background(work, image, bgdata);
    image_blend(image, newframe, profile_in, profile_out, bgdata);
    break;
  case BLEND_OP_SOURCE:
    get_image_background(work, image, bgdata);
    image_blend(image, newframe, profile_in, profile_out, bgdata);
    break;
  case BLEND_OP_OVER:
    image_blend(image, newframe, profile_in, profile_out, NULL);
    break;
  default:
    DEBUG_PRINT(" unknown blend_op type %d. why?\n", blend_op);
    break;
  }
  if (work->type != IMAGE_TYPE_INDEXED_RGBA_8) {
    blend_op = BLEND_OP_NONE;
  }
  image_set_blend_op(image, blend_op);
}

static int get_image_background(ImageWork *work, Image *image, unsigned char *pbgdata)
{
  BgPrefer bg_prefer;
  ImageType bgtype;
  ImageType type;
  CmsProfile *profile_in;
  CmsProfile *profile_out;
  int rc = FALSE;

  DEBUG_PRINT(" bg_prefer=%d bgtype=%d\n", work->bg_prefer, work->bgtype);
  type = image_get_type(image);
  bg_prefer = work->bg_prefer;
  if (bg_prefer == BG_USER) {
    switch (type) {
    case IMAGE_TYPE_INDEXED_RGB_8:
    case IMAGE_TYPE_INDEXED_RGBA_8:
      pbgdata[0] = work->bgdata[0];
      rc = TRUE;
      break;
    case IMAGE_TYPE_RGB_8:
      pbgdata[0] = work->bguser[0];
      pbgdata[1] = work->bguser[2];
      pbgdata[2] = work->bguser[4];
      rc = TRUE;
      break;
    case IMAGE_TYPE_RGB_16:
      memcpy(pbgdata, work->bguser, imageType_bytes_per_pixel(type));
      rc = TRUE;
      break;
    }
    return rc;
  }
  /* otherwise, bg_prefer = BG_IMAGE */
  bgtype = work->bgtype;
  if (bgtype == IMAGE_TYPE_UNKNOWN) {
    DEBUG_PRINT(" bgtype=%s\n", "unknown");
    return rc;
  }
  profile_in = imageWork_get_profile_in(work);
  profile_out = work->profile_out;
  if ((profile_in != NULL) && (profile_out != NULL)) {
    if (!IMAGE_TYPE_IS_INDEXED(bgtype)) {
      cms_data_transform(work->bgdata, bgtype, profile_in,
                         pbgdata, type, profile_out, 1);
      rc = TRUE;
    } else {
      unsigned char bgtmp[4];
      switch (type) {
      case IMAGE_TYPE_INDEXED_RGB_8:
      case IMAGE_TYPE_INDEXED_RGBA_8:
        pbgdata[0] = work->bgdata[0];
        rc = TRUE;
        break;
      case IMAGE_TYPE_RGB_8:
      case IMAGE_TYPE_RGB_16:
        /* indexed color to RGB */
        image_get_index(image, work->bgdata[0], bgtmp);
        cms_data_transform(bgtmp, IMAGE_TYPE_RGB_8, profile_in,
                           pbgdata, type, profile_out, 1);
        rc = TRUE;
        break;
      }
    }
    DEBUG_PRINT(" rc=%d bgtype=%d type=%d\n", rc, bgtype, type);
    return rc;
  }
  if ((type == bgtype) ||
      (IMAGE_TYPE_IS_INDEXED(type) && IMAGE_TYPE_IS_INDEXED(bgtype))) {
    /* same type */
    memcpy(pbgdata, work->bgdata, imageType_bytes_per_pixel(type));
    rc = TRUE;
  } else {
    void (*copy_data)(unsigned char *psrc, const unsigned char *pdst, int len);
    copy_data = NULL;
    switch (type) {
    case IMAGE_TYPE_RGB_8:
      if (bgtype == IMAGE_TYPE_GRAY_8) {
        copy_data = copy_gray8_to_rgb8;
      } else if (bgtype == IMAGE_TYPE_CMYK_8) {
        copy_data = copy_cmyk8_to_rgb8;
      }
      break;
    case IMAGE_TYPE_RGB_16:
      if (bgtype == IMAGE_TYPE_GRAY_16) {
        copy_data = copy_gray16_to_rgb16;
      }
      break;
    }
    if (copy_data != NULL) {
      (*copy_data)(pbgdata, work->bgdata, 1);
      rc = TRUE;
    }
  }
  DEBUG_PRINT(" rc=%d\n", rc);
  return rc;
}

int imageWork_get_dispose_op(ImageWork *work)
{
  return work->dispose_op;
}

Image *imageWork_get_image(ImageWork *work)
{
  return work->current;
}

void imageWork_get_updated_size(ImageWork *work, int *width, int *height, int *xoff, int *yoff)
{
  *xoff = work->updated_xoff;
  *yoff = work->updated_yoff;
  *width = work->updated_width;
  *height = work->updated_height;
}

void imageWork_set_profiles(ImageWork *work, CmsProfile *profile_in, CmsProfile *profile_out, int use_infile)
{
  work->profile_in = profile_in;
  work->profile_out = profile_out;
  work->use_infile = use_infile;
}

static CmsProfile *imageWork_get_profile_in(ImageWork *work)
{
  CmsProfile *profile;

  if (work->use_infile) {
    profile = work->profile_in;
  } else {
    profile = work->profile_embedded;
    if (profile == NULL) {
      profile = work->profile_in;
    }
  }
  return profile;
}

static void image_copy_data(Image *dst, Image *src)
{
  ImageType type_src, type_dst;
  unsigned char *pp;
  unsigned char *pdst;
  unsigned char *psrc;
  int w0, h0, w, h;
  int xoff, yoff;
  size_t bpp0, bpp; /* bytes per pixel */

  type_dst = image_get_type(dst);
  pdst = image_get_data(dst);
  image_get_size(dst, &w0, &h0);
  bpp = imageType_bytes_per_pixel(type_dst);
  type_src = image_get_type(src);
  psrc = image_get_data(src);
  image_get_size(src, &w, &h);
  image_get_offset(src, &xoff, &yoff);
  bpp0 = imageType_bytes_per_pixel(type_src);
  pp = pdst+(w0*yoff+xoff)*bpp0;
  if ((type_src == type_dst) ||
      (IMAGE_TYPE_IS_INDEXED(type_src) && IMAGE_TYPE_IS_INDEXED(type_dst))) {
    /* same type */
    if (w == w0) {
      memcpy(pp, psrc, w*h*bpp);
    } else {
      unsigned char *pp2;
      pp2 = psrc;
      for (int j = 0; j < h; j++) {
        memcpy(pp, pp2, w*bpp);
        pp += w0*bpp0;
        pp2 += w*bpp;
      }
    }
  } else {
    void (*copy_data)(unsigned char *pdst, const unsigned char *psrc, int len);
    copy_data = NULL;
    switch (type_dst) {
    case IMAGE_TYPE_RGB_8:
      if (type_src == IMAGE_TYPE_GRAY_8) {
        copy_data = copy_gray8_to_rgb8;
      } else if (type_src == IMAGE_TYPE_CMYK_8) {
        copy_data = copy_cmyk8_to_rgb8;
      }
      break;
    case IMAGE_TYPE_RGB_16:
      if (type_src == IMAGE_TYPE_GRAY_16) {
        copy_data = copy_gray16_to_rgb16;
      }
      break;
    }
    if (copy_data != NULL) {
      if (w == w0) {
        (*copy_data)(pp, psrc, w*h);
      } else {
        unsigned char *pp2;
        pp2 = psrc;
        for (int j = 0; j < h; j++) {
          (*copy_data)(pp, pp2, w);
          pp += w0*bpp0;
          pp2 += w*bpp;
        }
      }
    }
  }
}

static void image_copy_palette(Image *dst, Image *src)
{
  ImageType type_src, type_dst;
  int num;

  num = image_get_num_indexes(src);
  image_set_num_indexes(dst, num);
  type_dst = image_get_type(dst);
  type_src = image_get_type(src);
  if (type_src == type_dst) {
    for (int i = 0; i < num; i++) {
      unsigned char data[4];
      image_get_index(src, i, data);
      image_set_index(dst, i, data);
    }
  } else {
    void (*copy_data)(unsigned char *pdst, const unsigned char *psrc, int len);
    copy_data = NULL;
    switch (type_dst) {
    case IMAGE_TYPE_INDEXED_RGB_8:
      if (type_src == IMAGE_TYPE_INDEXED_GRAY_8) {
        copy_data = copy_gray8_to_rgb8;
      }
      break;
    case IMAGE_TYPE_INDEXED_RGBA_8:
      if (type_src == IMAGE_TYPE_INDEXED_GRAY_8) {
        copy_data = copy_gray8_to_rgba8;
      } else if (type_src == IMAGE_TYPE_INDEXED_GRAYA_8) {
        copy_data = copy_graya8_to_rgba8;
      } else if (type_src == IMAGE_TYPE_INDEXED_RGB_8) {
        copy_data = copy_rgb8_to_rgba8;
      }
      break;
    }
    if (copy_data != NULL) {
      for (int i = 0; i < num; i++) {
        unsigned char data_src[4], data_dst[4];
        image_get_index(src, i, data_src);
        (*copy_data)(data_dst, data_src, 1);
        image_set_index(dst, i, data_dst);
      }
    }
  }
}

static void image_fill_pixel(Image *image, unsigned char *pixel)
{
  ImageType type;
  unsigned char *pp;
  int w, h, bpp;

  type = image_get_type(image);
  image_get_size(image, &w, &h);
  pp = image_get_data(image);
  bpp = imageType_bytes_per_pixel(type);
  if (bpp == 1) {
    memset(pp, *pixel, w*h);
  } else if (bpp > 1) {
    for (int i = 0; i < w*h; i++) {
      for (int c = 0; c < bpp; c++) {
        *(pp++) = pixel[c];
      }
    }
  }
}

static void image_blend(Image *image, Image *newframe,
                        CmsProfile *profile_in, CmsProfile *profile_out,
                        unsigned char *bgdata)
{
  ImageType type, newtype;
  unsigned char *pp;
  unsigned char *pp1;
  unsigned char *pp2;
  int w0, h0, w, h;
  int xoff, yoff;

  pp1 = image_get_data(image);
  image_get_size(image, &w0, &h0);
  pp2 = image_get_data(newframe);
  image_get_size(newframe, &w, &h);
  image_get_offset(newframe, &xoff, &yoff);
  newtype = image_get_type(newframe);
  switch (newtype) {
  case IMAGE_TYPE_INDEXED_GRAY_8:
  case IMAGE_TYPE_INDEXED_GRAYA_8:
  case IMAGE_TYPE_INDEXED_RGB_8:
  case IMAGE_TYPE_INDEXED_RGBA_8:
    type = image_get_type(image);
    if (type == IMAGE_TYPE_RGB_8) {
      /* +indexed color to RGB8 */
      image_blend_index_to_rgb8(image, newframe, bgdata);
    } else {
      /* +indexed color to indexed color */
      image_copy_data(image, newframe);
    }
    break;
  case IMAGE_TYPE_GRAY_8:
  case IMAGE_TYPE_GRAY_16:
  case IMAGE_TYPE_RGB_8:
  case IMAGE_TYPE_RGB_16:
  case IMAGE_TYPE_CMYK_8:
    /* +GRAY or +RGB or +CMYK to RGB */
    if ((profile_in != NULL) && (profile_out != NULL)) {
      image_transform_data(newframe, profile_in, image, profile_out);
    } else {
      image_copy_data(image, newframe);
    }
    break;
  case IMAGE_TYPE_GRAYA_8:
  case IMAGE_TYPE_GRAYA_16:
  case IMAGE_TYPE_RGBA_8:
  case IMAGE_TYPE_RGBA_16:
    /* +GRAYA or +RGBA to RGB */
    if ((profile_in != NULL) && (profile_out != NULL)) {
      Image *tmpframe;
      tmpframe = image_transform_new(newframe, profile_in, profile_out);
      image_blend_rgb(image, tmpframe, bgdata);
      image_free(tmpframe);
    } else {
      image_blend_rgb(image, newframe, bgdata);
    }
    break;
  default:
    break;
  }
}

/*
 * +indexed color to RGB8.
 */
static void image_blend_index_to_rgb8(Image *image, Image *newframe, const unsigned char *bgdata)
{
  unsigned char *pp1;
  unsigned char *pp2;
  int w0, h0, w, h;
  int y, xoff, yoff;
  int bpp0;

  bpp0 = imageType_bytes_per_pixel(IMAGE_TYPE_RGB_8);
  pp1 = image_get_data(image);
  image_get_size(image, &w0, &h0);
  pp2 = image_get_data(newframe);
  image_get_size(newframe, &w, &h);
  image_get_offset(newframe, &xoff, &yoff);
  y = yoff;
  for (int j = 0; j < h; j++) {
    unsigned char *pp;
    pp = pp1+(w0*y+xoff)*bpp0;
    for (int i = 0; i < w; i++) {
      unsigned char tmp[4];
      image_get_index(newframe, *pp2, tmp);
      blend_c8(pp, tmp, bgdata, 3);
      pp += bpp0;
      pp2++;
    }
    y++;
  }
}

/*
 * +GRAYA or +RGBA to RGB.
 */
static void image_blend_rgb(Image *image, Image *newframe, const unsigned char *bgdata)
{
  void (*copy_data)(unsigned char *pdst, const unsigned char *psrc, int len);
  void (*blend_c)(unsigned char *p1, const unsigned char *p2, const unsigned char *bg, int num);
  ImageType type0;
  ImageType type;
  unsigned char *pp1;
  unsigned char *pp2;
  int bpp0, bpp; /* bytes per pixel */
  int w0, h0, w, h;
  int y, xoff, yoff;
  int num;

  type0 = image_get_type(image);
  type = image_get_type(newframe);
  switch (type) {
  case IMAGE_TYPE_GRAYA_8:
    copy_data = copy_graya8_to_rgba8;
    blend_c = blend_c8;
    break;
  case IMAGE_TYPE_GRAYA_16:
    copy_data = copy_graya16_to_rgba16;
    blend_c = blend_c16;
    break;
  case IMAGE_TYPE_RGBA_8:
    copy_data = NULL;
    blend_c = blend_c8;
    break;
  case IMAGE_TYPE_RGBA_16:
    copy_data = NULL;
    blend_c = blend_c16;
    break;
  default:
    return;
    break;
  }
  bpp0 = imageType_bytes_per_pixel(type0);
  bpp = imageType_bytes_per_pixel(type);
  pp1 = image_get_data(image);
  image_get_size(image, &w0, &h0);
  pp2 = image_get_data(newframe);
  image_get_size(newframe, &w, &h);
  image_get_offset(newframe, &xoff, &yoff);
  y = yoff;
  if (copy_data == NULL) {
    for (int j = 0; j < h; j++) {
      unsigned char *pp;
      pp = pp1+(w0*y+xoff)*bpp0;
      for (int i = 0; i < w; i++) {
        (*blend_c)(pp, pp2, bgdata, 3);
        pp += bpp0;
        pp2 += bpp;
      }
      y++;
    }
  } else {
    for (int j = 0; j < h; j++) {
      unsigned char *pp;
      pp = pp1+(w0*y+xoff)*bpp0;
      for (int i = 0; i < w; i++) {
        unsigned char tmp[8];
        (*copy_data)(tmp, pp2, 1);
        (*blend_c)(pp, tmp, bgdata, 3);
        pp += bpp0;
        pp2 += bpp;
      }
      y++;
    }
  }
}

static void copy_gray8_to_rgb8(unsigned char *pdst, const unsigned char *psrc, int len)
{
  for (int i = 0; i < len; i++) {
    *(pdst++) = *psrc;
    *(pdst++) = *psrc;
    *(pdst++) = *(psrc++);
  }
}

static void copy_gray8_to_rgba8(unsigned char *pdst, const unsigned char *psrc, int len)
{
  for (int i = 0; i < len; i++) {
    *(pdst++) = *psrc;
    *(pdst++) = *psrc;
    *(pdst++) = *(psrc++);
    *(pdst++) = 0xff;
  }
}

static void copy_gray16_to_rgb16(unsigned char *pdst, const unsigned char *psrc, int len)
{
  for (int i = 0; i < len; i++) {
    *(pdst++) = *psrc;
    *(pdst++) = *(psrc+1);
    *(pdst++) = *psrc;
    *(pdst++) = *(psrc+1);
    *(pdst++) = *(psrc++);
    *(pdst++) = *(psrc++);
  }
}

static void copy_graya8_to_rgba8(unsigned char *pdst, const unsigned char *psrc, int len)
{
  for (int i = 0; i < len; i++) {
    *(pdst++) = *psrc;
    *(pdst++) = *psrc;
    *(pdst++) = *(psrc++);
    *(pdst++) = *(psrc++);
  }
}

static void copy_graya16_to_rgba16(unsigned char *pdst, const unsigned char *psrc, int len)
{
  for (int i = 0; i < len; i++) {
    *(pdst++) = *psrc;
    *(pdst++) = *(psrc+1);
    *(pdst++) = *psrc;
    *(pdst++) = *(psrc+1);
    *(pdst++) = *psrc;
    *(pdst++) = *(psrc++);
    *(pdst++) = *(psrc++);
    *(pdst++) = *(psrc++);
  }
}

static void copy_rgb8_to_rgba8(unsigned char *pdst, const unsigned char *psrc, int len)
{
  for (int i = 0; i < len; i++) {
    *(pdst++) = *(psrc++);
    *(pdst++) = *(psrc++);
    *(pdst++) = *(psrc++);
    *(pdst++) = 0xff;
  }
}

static void copy_cmyk8_to_rgb8(unsigned char *pdst, const unsigned char *psrc, int len)
{
  for (int i = 0; i < len; i++) {
    unsigned int r_c, g_m, b_y, a_k;
    r_c = *(psrc++);
    g_m = *(psrc++);
    b_y = *(psrc++);
    a_k = *(psrc++);
    *(pdst++) = (r_c*a_k)/0xff;
    *(pdst++) = (g_m*a_k)/0xff;
    *(pdst++) = (b_y*a_k)/0xff;
  }
}

static void blend_c8(unsigned char *p1, const unsigned char *p2, const unsigned char *bg, int num)
{
  unsigned char *pp;
  unsigned char a;
  int bpp; /* bytes_per_pixel */

  bpp = sizeof(unsigned char)*num;
  a = *(p2+num);
  if (a == 0xff) {
    /* fully opaque */
    memcpy(p1, p2, bpp);
    return;
  }
  if (bg != NULL) {
    /* case BLEND_OP_SOURCE */
    if (a == 0x00) {
      /* fully transparent */
      memcpy(p1, bg, bpp);
      return;
    }
    if (memcmp(bg, p2, bpp) == 0) {
      /* same value */
      memcpy(p1, p2, bpp);
      return;
    }
    pp = p1;
    p1 = bg;
  } else {
    /* case BLEND_OP_OVER */
    if (a == 0x00) {
      /* fully transparent */
      return;
    }
    if (memcmp(p1, p2, bpp) == 0) {
      /* same value */
      return;
    }
    pp = p1;
  }
  for (int i = 0; i < num; i++) {
    unsigned int c1, c2, cc;
    unsigned char c;

    c1 = *(p1++);
    c2 = *(p2++);
    cc = (c1*(0xff-a)+c2*a)/0xff; /* = c1*(1.0-alpha)+c2*alpha */
    c = (unsigned char)cc;
    *(pp++) = c;
  }
}

static void blend_c16(unsigned char *p1, const unsigned char *p2, const unsigned char *bg, int num)
{
  unsigned char *pp;
  unsigned short a;
  int bpp; /* bytes_per_pixel */

  bpp = sizeof(unsigned char)*2*num;
  a = get_c16(p2+2*num);
  if (a == 0xffff) {
    /* fully opaque */
    memcpy(p1, p2, bpp);
    return;
  }
  if (bg != NULL) {
    /* case BLEND_OP_SOURCE */
    if (a == 0x0000) {
      /* fully transparent */
      memcpy(p1, bg, bpp);
      return;
    }
    if (memcmp(bg, p2, bpp) == 0) {
      /* same value */
      memcpy(p1, p2, bpp);
      return;
    }
    pp = p1;
    p1 = bg;
  } else {
    /* case BLEND_OP_OVER */
    if (a == 0x0000) {
      /* fully transparent */
      return;
    }
    if (memcmp(p1, p2, bpp) == 0) {
      /* same value */
      return;
    }
    pp = p1;
  }
  for (int i = 0; i < num; i++) {
    unsigned long c1, c2, cc;
    unsigned short c;

    c1 = get_c16(p1);
    c2 = get_c16(p2);
    cc = (c1*(0xffff-a)+c2*a)/0xffff; /* = c1*(1.0-alpha)+c2*a */
    c = (unsigned short)cc;
    set_c16(pp, c);
    p1 += 2;
    p2 += 2;
    pp += 2;
  }
}

static void get_rgb8(unsigned char *rgb, unsigned short *r, unsigned short *g, unsigned short *b)
{
  *r = get_c8(&rgb[0]);
  *g = get_c8(&rgb[1]);
  *b = get_c8(&rgb[2]);
}

static void get_rgb16(unsigned char *rgb, unsigned short *r, unsigned short *g, unsigned short *b)
{
  *r = get_c16(&rgb[0]);
  *g = get_c16(&rgb[2]);
  *b = get_c16(&rgb[4]);
}

static void set_rgb16(unsigned char *rgb, unsigned short r, unsigned short g, unsigned short b)
{
  set_c16(&rgb[0], r);
  set_c16(&rgb[2], g);
  set_c16(&rgb[4], b);
}

static unsigned short get_c16(unsigned char *c16)
{
  unsigned short c;

  c  = (c16[0] << 8) & 0xff00;
  c |= c16[1] & 0xff;
  return c;
}

static void set_c16(unsigned char *c16, unsigned short c)
{
  c16[0] = (c >> 8) & 0xff;
  c16[1] = c & 0xff;
}

static unsigned short get_c8(unsigned char *c8)
{
  unsigned short cc, c;

  cc = c8[0];
  c  = (cc << 8) & 0xff00;
  c |= cc & 0xff;
  return c;
}
