/* $Id$ */
/*
 * TIFF decoding.
 */
#include <stdio.h>
#include <stdlib.h>
#include <tiffio.h>
#include "module.h"
#include "cmsprofile.h"

int mod_read_TIFF(Entry *entry, ImageSet *images);
static Image *tiff_read(TIFF *tif);
static CmsProfile *get_cms_profile(TIFF *tif);
#if defined DEBUG
static void info_print(TIFF *tif);
#endif /* DEBUG */


int mod_read_TIFF(Entry *entry, ImageSet *images)
{
  TIFF *tif = NULL;
  int dircount = 0;
  int width, height;

  tif = TIFFOpen(entry->filename, "r");
  if (tif == NULL) {
    fprintf(stderr, "can't open %s\n", entry->filename);
    return FALSE;
  }
  width = height = -1;
  do {
    Image *image;

    image = tiff_read(tif);
    if (image != NULL) {
      int w, h;
      image_get_size(image, &w, &h);
      if (w > width) {
        width = w;
      }
      if (h > height) {
        height = h;
      }
      imageSet_add_image(images, image);
    }
#if defined DEBUG
    fprintf(stderr, "tiff: image[%d]\n", dircount);
#endif /* DEBUG */
    dircount++;
  } while (TIFFReadDirectory(tif));
  TIFFClose(tif);

  if ((width > 0) && (height > 0)) {
    imageSet_set_size(images, width, height);
  }
  return TRUE;
}

static Image *tiff_read(TIFF *tif)
{
  uint32 *raster;
  uint32 width, height;
  uint16 photometric;
  uint16 bps, spp, extra;
  uint16 *info;
  unsigned char *src, *dst;
  unsigned char *data;

  if (!TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &width)) {
    TIFFError(TIFFFileName(tif), "Image does not define its width");
    return NULL;
  }
  if (!TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &height)) {
    TIFFError(TIFFFileName(tif), "Image does not define its height");
    return NULL;
  }
#if defined DEBUG
  info_print(tif);
#endif /* DEBUG */
  TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, &photometric);
  TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &bps);
  TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &spp);
  TIFFGetField(tif, TIFFTAG_EXTRASAMPLES, &extra, &info);
  switch (photometric) {
  case PHOTOMETRIC_MINISWHITE:
  case PHOTOMETRIC_MINISBLACK:
    /* GRAY_8 */
    break;
  case PHOTOMETRIC_RGB:
    /* RGB_8 */
    break;
  case PHOTOMETRIC_SEPARATED:
    if (spp == 4) {
      /* CMYK_8 */
    } else if (spp == 3) {
      /* CMY_8 */
    }
    break;
  default:
    break;
  }
  raster = (uint32*)_TIFFmalloc(width*height*sizeof(uint32));
  if (raster == NULL) {
    TIFFError(TIFFFileName(tif), "No space for raster buffer");
    return NULL;
  }
  /* Read the image in one chunk into an RGBA array */
  if (!TIFFReadRGBAImageOriented(tif, width, height, raster, ORIENTATION_TOPLEFT, 0)) {
    TIFFError(TIFFFileName(tif), "TIFFReadRGBAImage not supported this image");
    _TIFFfree(raster);
    return NULL;
  }
  data = (unsigned char *)malloc(width*height*imageType_bytes_per_pixel(IMAGE_TYPE_RGBA_8));
  src = (unsigned char *)raster;
  dst = data;
  for (uint32 row = 0; row < height; row++) {
    for (uint32 col = 0; col < width; col++) {
      *(dst++) = *(src++); /* red */
      *(dst++) = *(src++); /* green */
      *(dst++) = *(src++); /* blue */
      *(dst++) = *(src++); /* alpha */
    }
  }
  _TIFFfree(raster);

  Image *image;
  image = image_new();
  image_set_data(image, data);
  image_set_size(image, width, height);
  image_set_type(image, IMAGE_TYPE_RGBA_8);
  return image;
}

/*
 * This snippet shows how to deal with Embedded profiles in TIFF
 * files using Sam Leffer's library.
 */
static CmsProfile *get_cms_profile(TIFF *tif)
{
  CmsProfile *profile;
  void *buffer;
  uint32 len;
  float *chr;

  if (TIFFGetField(tif, TIFFTAG_ICCPROFILE, &len, &buffer)) {
    profile = cmsProfile_open_mem(buffer, len);
    return profile;
  }
  if (TIFFGetField(tif, TIFFTAG_PRIMARYCHROMATICITIES, &chr)) {
    float *white;
    double red_x, red_y;
    double green_x, green_y;
    double blue_x, blue_y;
    red_x = chr[0];
    red_y = chr[1];
    green_x = chr[2];
    green_y = chr[3];
    blue_x = chr[4];
    blue_y = chr[5];
    if (TIFFGetField(tif, TIFFTAG_WHITEPOINT, &white)) {
      unsigned short *gm_r, *gm_g, *gm_b;
      double white_x, white_y;
      white_x = white[0];
      white_y = white[1];
      if (TIFFGetField(tif, TIFFTAG_TRANSFERFUNCTION, &gm_r, &gm_g, &gm_b)) {
        profile = NULL;
        return profile;
      }
      return NULL;
    }
    return NULL;
  }
  return NULL;
}

#if defined DEBUG
static void info_print(TIFF *tif)
{
  uint32 width, height;
  uint16* sampleinfo;
  uint16 extrasamples;
  uint16 photometric, samplesperpixel, bitspersample;
  CmsProfile *profile;
  const char *str;

  if (!TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &width)) {
    TIFFError(TIFFFileName(tif), "Image does not define its width");
  }
  if (!TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &height)) {
    TIFFError(TIFFFileName(tif), "Image does not define its height");
  }
  if (!TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, &photometric)) {
    TIFFError(TIFFFileName(tif), "Image does not define its photometric");
  }
  if (!TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &samplesperpixel)) {
    TIFFError(TIFFFileName(tif), "Image does not define its sampleperpixel");
  }
  if (!TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &bitspersample)) {
    TIFFError(TIFFFileName(tif), "Image does not define its bitspersample");
  }
  extrasamples = 0;
  if (!TIFFGetField(tif, TIFFTAG_EXTRASAMPLES, &extrasamples, &sampleinfo)) {
    TIFFError(TIFFFileName(tif), "Image does not define its extrasamples");
  }
  switch (photometric) {
  case PHOTOMETRIC_MINISWHITE:
    str = "MINISWHITE";
    break;
  case PHOTOMETRIC_MINISBLACK:
    str = "MINISBLACK";
    break;
  case PHOTOMETRIC_RGB: 
    str = "RGB";
    break;
  case PHOTOMETRIC_PALETTE:
    str = "PALETTE";
    break;
  case PHOTOMETRIC_MASK:
    str = "MASK";
    break;
  case PHOTOMETRIC_SEPARATED:
    str = "SEPARATED";
    break;
  case PHOTOMETRIC_YCBCR:
    str = "YCBCR";
    break;
  case PHOTOMETRIC_CIELAB:
    str = "CIELAB";
    break;
  case PHOTOMETRIC_ICCLAB:
    str = "ICCLAB";
    break;
  case PHOTOMETRIC_ITULAB:
    str = "ITULAB";
    break;
  case PHOTOMETRIC_LOGL:
    str = "LOGL";
    break;
  case PHOTOMETRIC_LOGLUV:
    str = "LOGLUV";
  default:
    str = "unknown";
    break;
  }
  fprintf(stderr,
          "tiff: (%dx%d) photometric: %s, sample/pixel: %u, bits/sample: %u\n",
          width, height, str, samplesperpixel, bitspersample);
  if (extrasamples > 0) {
    fprintf(stderr, "tiff: extrasamples=%u\n", extrasamples);
    switch (sampleinfo[0]) {
    case EXTRASAMPLE_UNSPECIFIED: /* workaround for some images without */
      fprintf(stderr, "tiff: extrasample: %s\n", "UNSPECIFIED");
      break;
    case EXTRASAMPLE_ASSOCALPHA: /* data is pre-multiplied */
      fprintf(stderr, "tiff: extrasample: %s\n", "ASSOCALPHA");
      break;
    case EXTRASAMPLE_UNASSALPHA: /* data is not pre-multiplied */
      fprintf(stderr, "tiff: extrasample: %s\n", "UNASSALPHA");
      break;
    }
  }
}
#endif /* DEBUG */
